from tequila.circuit import gates
from numpy import pi
import typing
from numbers import Real
from tequila import Variable,QCircuit
from copy import deepcopy




class FermionicGateImpl(gates.QubitExcitationImpl):
    """
        Small helper class for Fermionic Excictation Gates
        Mainly so that "FermionicGate is displayed when circuits are printed
    """
    def __init__(self, generator, p0, transformation,select=None ,indices=None,condense=False,up_then_down=False,two_qubit=False,*args, **kwargs):
        if "opt" in kwargs:
            self.opt = kwargs["opt"]
            kwargs.pop("opt")
        else: self.opt =True
        super().__init__(generator=generator, target=generator.qubits, p0=p0, *args, **kwargs)
        self._name = "FermionicExcitation"
        self.transformation = transformation
        self.indices = indices
        self.select = select
        self.condense = condense
        self.up_then_down = up_then_down
        self.two_qubit = two_qubit
        self.n_orbitals = len(select)
        self.FER_SO, self.pos = self.select_to_list()
        if not hasattr(indices[0], "__len__"):
            self.indices = [(indices[2 * i], indices[2 * i + 1]) for i in range(len(indices) // 2)]
        self.sign = self.format_excitation_variables(self.indices)
        self.indices = self.format_excitation_indices(self.indices)
    def fermionic_excitation(self, angle: typing.Union[Real, Variable, typing.Hashable], indices: typing.List,
                             control: typing.Union[int, typing.List] = None,) -> QCircuit:
        '''
            Excitation [(i,j),(k,l)],... compiled following https://doi.org/10.1103/PhysRevA.102.062612
            opt: whether to optimized CNOT H CNOT --> Rz Rz CNOT Rz
        '''
        lto = []
        lfrom = []
        if isinstance(angle, str) or isinstance(angle, tuple):
            angle = Variable(angle)
        if not hasattr(indices[0],"__len__"):
            indices = [(indices[2 * i], indices[2 * i + 1]) for i in range(len(indices) // 2)]
        for pair in indices:
            if self.select is None or self.select[pair[0] // 2] == "F" or (self.select[pair[0] // 2] == "B" and not pair[0] % 2) or self.two_qubit:
                lfrom.append(self.pos[pair[0]])
            if self.select is None or self.select[pair[1] // 2] == "F" or (self.select[pair[1] // 2] == "B" and not pair[1] % 2) or self.two_qubit:
                lto.append(self.pos[pair[1]])
        Upair = QCircuit()
        for i in range(len(lfrom) - 1):
            Upair += gates.CNOT(lfrom[i + 1], lfrom[i]) + gates.X(lfrom[i])
        for i in range(len(lto) - 1):
            Upair += gates.CNOT(lto[i + 1], lto[i]) + gates.X(lto[i])
        Upair += gates.CNOT(lto[-1], lfrom[-1])
        crt = lfrom[::-1] + lto
        Uladder = QCircuit()
        pairs = lfrom + lto
        pairs.sort()
        idx = []
        for pair in indices:
            idx.append(pair[0])
            idx.append(pair[1])
        idx.sort()
        orbs = []
        for o in range(len(idx) // 2):
            orbs += [*range(idx[2 * o] + 1, idx[2 * o + 1])]
        orbs = [self.pos[i] for i in orbs if self.select[i//2]=="F"]
        if len(orbs):
            for o in range(len(orbs) - 1):
                Uladder += gates.CNOT(orbs[o], orbs[o + 1])
            Uladder += gates.CZ(orbs[-1], lto[-1])
        crt.pop(-1)
        if control is not None and (isinstance(control, int) or len(control) == 1):
            if isinstance(control, int):
                crt.append(control)
            else:
                crt = crt + control
            control = []
        Ur = self.cCRy(target=lto[-1], dcontrol=crt, angle=angle, control=control)
        Upair2 = Upair.dagger()
        if self.opt:
            Ur.gates.pop(-1)
            Ur.gates.pop(-1)
            Upair2.gates.pop(0)
            Ur += gates.Rz(pi / 2, target=lto[-1]) + gates.Rz(-pi / 2, target=lfrom[-1])
            Ur += gates.CNOT(lto[-1], lfrom[-1]) + gates.Rz(pi / 2, target=lfrom[-1]) + gates.H(lfrom[-1])
        return Upair + Uladder + Ur + Uladder.dagger() + Upair2

    def compile(self, *args, **kwargs):
        if self.is_convertable_to_qubit_excitation():
            target = []
            for x in self.indices:
                for y in x:
                    target.append(y)
            U = gates.QubitExcitation(target=target, angle=self.parameter, control=self.control)
        else:
            if self.transformation.lower().strip("_") == "jordanwigner":
                U =  self.fermionic_excitation(angle=self.sign*self.parameter, indices=self.indices, control=self.control)
            else:
                U = gates.Trotterized(generator=self.generator, control=self.control, angle=self.parameter, steps=1)
        return U

    def format_excitation_indices(self, idx):
        """
        Consistent formatting of excitation indices
        idx = [(p0,q0),(p1,q1),...,(pn,qn)]
        sorted as: p0<p1<pn and pi<qi
        :param idx: list of index tuples describing a single(!) fermionic excitation
        :return: list of index tuples
        """

        idx = [tuple(sorted(x)) for x in idx]
        idx = sorted(idx, key=lambda x: x[0])
        return list(idx)

    def format_excitation_variables(self, idx):
        """
        Consistent formatting of excitation variable
        idx = [(p0,q0),(p1,q1),...,(pn,qn)]
        sorted as: pi<qi and p0 < p1 < p2
        :param idx: list of index tuples describing a single(!) fermionic excitation
        :return: sign of the variable with re-ordered indices
        """
        sig = 1
        for pair in idx:
            if pair[1] > pair[0]:
                sig *= -1
        for pair in range(len(idx) - 1):
            if idx[pair + 1][0] > idx[pair][0]:
                sig *= -1
        return sig

    def cCRy(self, target: int, dcontrol: typing.Union[list, int], control: typing.Union[list, int],
             angle: typing.Union[Real, Variable, typing.Hashable], case: int = 1) -> QCircuit:
        '''
        Compilation of CRy as on https://doi.org/10.1103/PhysRevA.102.062612
        If not control passed, Ry returned
        Parameters
        ----------
        case: if 1 employs eq. 12 from the paper, if 0 eq. 13
        '''
        if control is not None and not len(control):
            control = None
        if isinstance(dcontrol, int):
            dcontrol = [dcontrol]
        if not len(dcontrol):
            return gates.Ry(angle=angle, target=target, control=control)
        else:
            if isinstance(angle, str):
                angle = Variable(angle)
            U = QCircuit()
            aux = dcontrol[0]
            ctr = deepcopy(dcontrol)
            ctr.pop(0)
            if case:
                U += self.cCRy(target=target, dcontrol=ctr, angle=angle / 2, case=1, control=control) + gates.H(aux) + gates.CNOT(target, aux)
                U += self.cCRy(target=target, dcontrol=ctr, angle=-angle / 2, case=0, control=control) + gates.CNOT(target, aux) + gates.H(aux)
            else:
                U += gates.H(aux) + gates.CNOT(target, aux) + self.cCRy(target=target, dcontrol=ctr, angle=-angle / 2,case=0, control=control)
                U += gates.CNOT(target, aux) + gates.H(aux) + self.cCRy(target=target, dcontrol=ctr, angle=angle / 2, case=1, control=control)
            return U
    def __str(self):
        if self.indices is not None:
            return "FermionicExcitation({})".format(str(self.indices))
        return "FermionicExcitation"

    def __repr__(self):
        return self.__str__()
    def is_convertable_to_qubit_excitation(self):
        """
        spin-paired double excitations (both electrons occupy the same spatial orbital and are excited to another spatial orbital)
        in the jordan-wigner representation are identical to 4-qubit excitations which can be compiled more efficient
        this function hels to automatically detect those cases
        Returns
        -------

        """
        return False
        if not self.transformation.lower().strip("_") == "jordanwigner": return False
        if not len(self.indices) == 2: return False
        if not self.indices[0][0] // 2 == self.indices[1][0] // 2: return False
        if not self.indices[0][1] // 2 == self.indices[1][1] // 2: return False
        return True

    def select_to_list(self):
        """
        Internal function
        Read the select string to make the proper Fer and Bos lists
        :return : list of MOs for the Bos, MOs and SOs for the Fer space
        """

        hcb = 0
        FER_SO = []
        sel = self.select
        pos = {}
        up = self.up_then_down
        two = self.two_qubit
        for i in sel:
            if (sel[i] == "B"):
                pos[2 * i] = i + (i - hcb) * (not up)
                if two:
                    pos[2 * i + 1] = i + self.n_orbitals * up + (not up) * (i + 1)
                    FER_SO.append(pos[2 * i])
                    FER_SO.append(pos[2 * i + 1])
                elif self.condense:
                    hcb += 1
            else:
                pos[2 * i] = i + (i - hcb) * (not up)
                pos[2 * i + 1] = i - hcb + up * self.n_orbitals + (not up) * (i + 1)
                FER_SO.append(pos[2 * i])
                FER_SO.append(pos[2 * i + 1])
        FER_SO.sort()
        return FER_SO, pos


import openfermion
from sympy.codegen.rewriting import powm1_opt
from tequila import TequilaException,QCircuit,Objective,Molecule
from tequila.quantumchemistry.qc_base import QuantumChemistryBase
import typing
import numpy as np
from sunrise.fermionic_operations.gates import *
from sunrise.fermionic_operations.circuit import FCircuit
from typing import Union

def make_fermionic_hamiltonian(one_body_integrals, two_body_integrals, constant, *args, **kwargs):

    one_body_coefficients, two_body_coefficients = openfermion.chem.molecular_data.spinorb_from_spatial(
        one_body_integrals, two_body_integrals)

    molecular_hamiltonian = openfermion.ops.representations.InteractionOperator(
        constant, one_body_coefficients, 1 / 2 * two_body_coefficients)
    fop = openfermion.transforms.get_fermion_operator(molecular_hamiltonian)

    return fop


def make_excitation_generator_op(indices: typing.Iterable[typing.Tuple[int, int]], form: str = 'fermionic')-> openfermion.FermionOperator:
    """
    Notes
    ----------
    Creates the transformed hermitian generator of UCC type unitaries:
          M(a^\dagger_{a_0} a_{i_0} a^\dagger{a_1}a_{i_1} ... - h.c.)
          where the qubit map M depends is self.transformation

    Parameters
    ----------
    indices : typing.Iterable[typing.Tuple[int, int]] :
        List of tuples [(a_0, i_0), (a_1, i_1), ... ] - recommended format, in spin-orbital notation (alpha odd numbers, beta even numbers)
        can also be given as one big list: [a_0, i_0, a_1, i_1 ...]
    form : str : (Default value None):
        Manipulate the generator to involution or projector
        set form='involution' or 'projector'
        the default is no manipulation which gives the standard fermionic excitation operator back
    Returns
    -------
    type
        1j*Transformed qubit excitation operator, depends on self.transformation
    """


    # check indices and convert to list of tuples if necessary
    if len(indices) == 0:
        raise TequilaException("make_excitation_operator: no indices given")
    elif not isinstance(indices[0], typing.Iterable):
        if len(indices) % 2 != 0:
            raise TequilaException("make_excitation_generator: unexpected input format of indices\n"
                                   "use list of tuples as [(a_0, i_0),(a_1, i_1) ...]\n"
                                   "or list as [a_0, i_0, a_1, i_1, ... ]\n"
                                   "you gave: {}".format(indices))
        converted = [(indices[2 * i], indices[2 * i + 1]) for i in range(len(indices) // 2)]
    else:
        converted = indices

    # convert everything to native python int
    # otherwise openfermion will complain
    converted = [(int(pair[0]), int(pair[1])) for pair in converted]

    # convert to openfermion input format
    ofi = []
    dag = []
    for pair in converted:
        assert (len(pair) == 2)
        ofi += [(int(pair[0]), 1),
                (int(pair[1]), 0)]  # openfermion does not take other types of integers like numpy.int64
        dag += [(int(pair[0]), 0), (int(pair[1]), 1)]
        if pair[0] == pair[1]:
            number_op = True
        else: 
            number_op = False
    op = openfermion.FermionOperator(tuple(ofi), -1.j)  # 1j makes it hermitian
    op += openfermion.FermionOperator(tuple(reversed(dag)), 1.j)

    if isinstance(form, str) and form.lower() != 'fermionic':
        # indices for all the Na operators
        Na = [x for pair in converted for x in [(pair[0], 1), (pair[0], 0)]]
        # indices for all the Ma operators (Ma = 1 - Na)
        Ma = [x for pair in converted for x in [(pair[0], 0), (pair[0], 1)]]
        # indices for all the Ni operators
        Ni = [x for pair in converted for x in [(pair[1], 1), (pair[1], 0)]]
        # indices for all the Mi operators
        Mi = [x for pair in converted for x in [(pair[1], 0), (pair[1], 1)]]

        # can gaussianize as projector or as involution (last is default)
        if form.lower() == "p+":
            op *= 0.5
            op += openfermion.FermionOperator(Na + Mi, 0.5)
            op += openfermion.FermionOperator(Ni + Ma, 0.5)
        elif form.lower() == "p-":
            op *= 0.5
            op += openfermion.FermionOperator(Na + Mi, -0.5)
            op += openfermion.FermionOperator(Ni + Ma, -0.5)

        elif form.lower() == "g+":
            op += openfermion.FermionOperator([], 1.0)  # Just for clarity will be subtracted anyway
            op += openfermion.FermionOperator(Na + Mi, -1.0)
            op += openfermion.FermionOperator(Ni + Ma, -1.0)
        elif form.lower() == "g-":
            op += openfermion.FermionOperator([], -1.0)  # Just for clarity will be subtracted anyway
            op += openfermion.FermionOperator(Na + Mi, 1.0)
            op += openfermion.FermionOperator(Ni + Ma, 1.0)
        elif form.lower() == "p0":
            # P0: we only construct P0 and don't keep the original generator
            # op = openfermion.FermionOperator([], 1.0)  # Just for clarity will be subtracted anyway <- biggest lie i have ever heard it is useless and destroys everything why the fuck is this there
            op = openfermion.FermionOperator(Na + Mi, -1.0)
            op += openfermion.FermionOperator(Ni + Ma, -1.0)
        else:
            raise TequilaException(
                "Unknown generator form {}, supported are G, P+, P-, G+, G- and P0".format(form))
    
    if number_op:
        op = openfermion.FermionOperator(tuple(ofi), 1)
      
    return op


def rotate_molecule(U:FCircuit,mol:QuantumChemistryBase,angles:Union[dict,None]=None)->QuantumChemistryBase:
    """
    U: a FCircuit with ONLY UR gates
    mol: Tequila Molecule you want to rotate
    angles: dictionary to map the circuit variables in case they are not mapped yet
    Output: rotated molecule
    """
    #TODO: at some point this function should go inside the qcbase, but better when implemented on tequila
    U = U.to_udud(norb=mol.n_orbitals)
    if angles is not None:
        U = U.map_variables(angles)
    core = 0
    while mol.integral_manager.orbitals[core].idx is None:
        core += 1
    n_mo = len(mol.integral_manager.orbital_coefficients)
    params = []
    indices = []
    for gate in reversed(U.gates):
        assert gate._name == 'UR'
        # Extract parameters and indices from gates
        params.append(gate.variables)
        g = []
        for idx in gate.indices[0]: #we only need the first index of each UR
            g.extend((i+2*core)//2 for i in idx)
        indices.append(g)
    # Create a matrix for each UR gate
    rot_matrices = []
    for k,index in enumerate(indices):
        tmp = np.eye(n_mo)
        if isinstance(params[k], Objective):
            params[k] = params[k]()
        tmp[index[0]][index[0]] = np.cos(params[k]/2)
        tmp[index[0]][index[1]] = np.sin(params[k]/2)
        tmp[index[1]][index[0]] = -np.sin(params[k]/2)
        tmp[index[1]][index[1]] = np.cos(params[k]/2)
        rot_matrices.append(tmp)

    # Multiply all matrices
    tmatrix = rot_matrices[0]
    for matrix in rot_matrices[1:]:
        tmatrix = np.dot(tmatrix,matrix)
    mol.integral_manager.transform_orbitals(tmatrix.T)
    return mol



from tencirchem.static.evolve_civector import *
from tencirchem.static.evolve_civector import _get_gradients_civector,_get_gradients_civector_nocache
from tequila import simulate,grad,QTensor
from tequila.objective.objective import Objective,Variable,FixedVariable
import tensorcircuit as tc
from typing import  Union
from numpy import zeros

def get_expval_and_grad_civector(
    angles, hamiltonian, n_qubits, n_elec_s,total_variables,
    params,ex_ops: Tuple, mode: str = "fermion", init_state=None,
    params_bra=None,ex_ops_bra:Union[Tuple,None]=None,init_state_bra=None):
    #Fist check bra
    if ex_ops_bra is None: ex_ops_bra = ex_ops
    if params_bra is None: params_bra = params
    if init_state_bra is None: init_state_bra = init_state
    #We receive list [x0,x1,...] refering to the total variables
    assert len(angles)==len(total_variables)
    #Then we create {v0:x0,v1:x1,...}
    dangles = {total_variables[i].name:angles[i] for i in range(len(angles))}
    #To build the CIVector we need the already maped params
    pa = []
    for p in params:
        pa.append(map_variables(p,dangles))
    pa_bra = []
    for p in params:
        pa_bra.append(map_variables(p,dangles))
    map_params = tc.backend.numpy(tc.backend.convert_to_tensor(pa).astype(tc.rdtypestr))
    map_params_bra = tc.backend.numpy(tc.backend.convert_to_tensor(pa_bra).astype(tc.rdtypestr))
    ket = get_civector(map_params    ,  n_qubits, n_elec_s, ex_ops    , [*range(len(ex_ops))]    , mode, init_state)
    bra = get_civector(map_params_bra, n_qubits , n_elec_s, ex_ops_bra, [*range(len(ex_ops_bra))], mode, init_state_bra)
    
    #Compute Hvects and energy 
    hbra = tc.backend.numpy(apply_op(hamiltonian, bra))
    hket = tc.backend.numpy(apply_op(hamiltonian, ket))
    energy = hbra @ ket
    #We got the gradient with respect to each ex_op, since assumed each ex_ops has different theta 
    # already cached
    op_tensors = get_operator_tensors(n_qubits, n_elec_s, ex_ops, mode=mode)
    op_tensors_bra = get_operator_tensors(n_qubits, n_elec_s, ex_ops_bra, mode=mode)

    theta_tensors = get_theta_tensors(map_params, [*range(len(ex_ops))])
    theta_tensors_bra = get_theta_tensors(map_params_bra, [*range(len(ex_ops_bra))])
    op_tensors = list(op_tensors) + list(theta_tensors)
    op_tensors_bra = list(op_tensors_bra) + list(theta_tensors_bra)
    
    
    gradients_beforesum = _get_gradients_civector(hbra, ket, *op_tensors[1:])
    gradients_beforesum_bra = _get_gradients_civector(hket, bra, *op_tensors_bra[1:])

    gradients_beforesum = tc.backend.numpy(gradients_beforesum)
    gradients_beforesum_bra = tc.backend.numpy(gradients_beforesum_bra)
    #Now we have to compute the gradient of all parameters with respect to all possible variables
    ang_grad = np.zeros((len(params),len(total_variables)))
    ang_grad_bra = np.zeros((len(params_bra),len(total_variables)))
    #TODO: Improve this using tequila Objective, and copy paste in the other functions
    for i,pa in enumerate(params):
        if isinstance(pa,FixedVariable):
            continue
        for j,an in enumerate(total_variables):
            ang_grad[i,j]=simulate(grad(1*pa,an),variables=dangles) 
    for i,pa in enumerate(params_bra):
        if isinstance(pa,FixedVariable):
            continue
        for j,an in enumerate(total_variables):
            ang_grad_bra[i,j]=simulate(grad(1*pa,an),variables=dangles)
    gradients = np.add(gradients_beforesum.dot(ang_grad),gradients_beforesum_bra.dot(ang_grad_bra))

    return energy,  gradients, bra @ ket

def get_energy_and_grad_civector(
    angles, hamiltonian, n_qubits, n_elec_s, total_variables,
    params, ex_ops: Tuple,  mode: str = "fermion", init_state=None):
    
    assert len(angles)==len(total_variables)
    dangles = {total_variables[i].name:angles[i] for i in range(len(angles))}
    pa = []
    for p in params: 
        pa.append(map_variables(p,dangles))
    map_params = tc.backend.numpy(tc.backend.convert_to_tensor(pa).astype(tc.rdtypestr))
    ket = get_civector(map_params, n_qubits, n_elec_s, ex_ops, [*range(len(ex_ops))], mode, init_state)
    bra = apply_op(hamiltonian, ket)
    energy = bra @ ket
    # already cached
    op_tensors = get_operator_tensors(n_qubits, n_elec_s, ex_ops, mode=mode)
    theta_tensors = get_theta_tensors(map_params, [*range(len(ex_ops))])
    op_tensors = list(op_tensors) + list(theta_tensors)
    gradients_beforesum = _get_gradients_civector(bra, ket, *op_tensors[1:])
    gradients_beforesum = tc.backend.numpy(gradients_beforesum)
    ang_grad = np.zeros((len(params),len(total_variables)))
    for i,pa in enumerate(params):
        if isinstance(pa,FixedVariable):
            continue
        for j,an in enumerate(total_variables):
            ang_grad[i,j]=simulate(grad(1*pa,an),variables=dangles)
    gradients = gradients_beforesum.dot(ang_grad)
    return energy, 2 * gradients

def get_expval_and_grad_civector_nocache(
    angles, hamiltonian, n_qubits, n_elec_s,total_variables,
    params,ex_ops: Tuple, mode: str = "fermion", init_state=None,
    params_bra=None,ex_ops_bra:Tuple=None,init_state_bra=None):
    
    if ex_ops_bra is None: ex_ops_bra = ex_ops
    if params_bra is None: params_bra = params
    if init_state_bra is None: init_state_bra = init_state
    assert len(angles)==len(total_variables)
    #Then we create {v0:x0,v1:x1,...}
    dangles = {total_variables[i].name:angles[i] for i in range(len(angles))}
    #To build the CIVector we need the already maped params
    pa = []
    for p in params:
        pa.append(map_variables(p,dangles))
    pa_bra = []
    for p in params:
        pa_bra.append(map_variables(p,dangles))
    map_params = tc.backend.numpy(tc.backend.convert_to_tensor(pa).astype(tc.rdtypestr))
    map_params_bra = tc.backend.numpy(tc.backend.convert_to_tensor(pa_bra).astype(tc.rdtypestr))
    ket = get_civector_nocache(map_params, n_qubits, n_elec_s, ex_ops, [*range(len(ex_ops))], mode, init_state)
    bra = get_civector_nocache(map_params_bra, n_qubits, n_elec_s, ex_ops_bra, [*range(len(ex_ops_bra))], mode, init_state_bra)
    hbra = apply_op(hamiltonian, bra)
    hket = apply_op(hamiltonian, ket)
    energy = hbra @ ket

    gradients_beforesum = _get_gradients_civector_nocache(hbra, ket, map_params, n_qubits, n_elec_s, ex_ops, [*range(len(ex_ops))], mode)
    gradients_beforesum = tc.backend.numpy(gradients_beforesum)
    gradients_beforesum_bra = _get_gradients_civector_nocache(hket, bra, map_params_bra, n_qubits, n_elec_s, ex_ops_bra, [*range(len(ex_ops_bra))], mode)
    gradients_beforesum_bra = tc.backend.numpy(gradients_beforesum_bra)

    ang_grad = np.zeros((len(params),len(total_variables)))
    ang_grad_bra = np.zeros((len(params_bra),len(total_variables)))
    for i,pa in enumerate(params):
        if isinstance(pa,FixedVariable):
            continue
        for j,an in enumerate(total_variables):
            ang_grad[i,j]=simulate(grad(1*pa,an),variables=dangles) 
    for i,pa in enumerate(params_bra):
        if isinstance(pa,FixedVariable):
            continue
        for j,an in enumerate(total_variables):
            ang_grad_bra[i,j]=simulate(grad(1*pa,an),variables=dangles)
    gradients = np.add(gradients_beforesum.dot(ang_grad),gradients_beforesum_bra.dot(ang_grad_bra))

    return energy, gradients, bra @ ket

def get_energy_and_grad_civector_nocache(
    angles, hamiltonian, n_qubits, n_elec_s, total_variables,
    params, ex_ops: Tuple,  mode: str = "fermion", init_state=None):

    assert len(angles)==len(total_variables)
    dangles = {total_variables[i].name:angles[i] for i in range(len(angles))}
    pa = []
    for p in params:
        pa.append(map_variables(p,dangles))
    map_params = tc.backend.numpy(tc.backend.convert_to_tensor(pa).astype(tc.rdtypestr))

    ket = get_civector_nocache(map_params, n_qubits, n_elec_s, ex_ops, [*range(len(ex_ops))], mode, init_state)
    bra = apply_op(hamiltonian, ket)
    energy = bra @ ket

    gradients_beforesum = _get_gradients_civector_nocache(bra, ket, map_params, n_qubits, n_elec_s, ex_ops, [*range(len(ex_ops))], mode)
    gradients_beforesum = tc.backend.numpy(gradients_beforesum)

    ang_grad = np.zeros((len(params),len(total_variables)))
    for i,pa in enumerate(params):
        if isinstance(pa,FixedVariable):
            continue
        for j,an in enumerate(total_variables):
            ang_grad[i,j]=simulate(grad(1*pa,an),variables=dangles)
    gradients = gradients_beforesum.dot(ang_grad)
    
    return energy, 2 * gradients

def map_variables(x:Union[Variable,Objective],dvariables:dict):
    if isinstance(x,Variable):
        x = x.map_variables(dvariables)
    elif isinstance(x,Objective):
        x=simulate(x,dvariables)
    return x


import copy
from typing import List, Optional

import tequila as tq

from ..core import CircuitState, CircuitStyle
from ..core import Gate


# TODO: add documentation, since I did not create this, I am unsure what it actually does and more importantly why
class GenericGate(Gate):
    """
    This is a generic gate, which renders the gate according to its name and QCircuit
    Trotter gates are left just as U.export_to(name.qpic), take care of the qubits yourself
    """


    # TODO: maybe something can be done with the Trotter gates?
    def __init__(self, U: tq.QCircuit, name: Optional[str] =None, n_qubits_is_double: bool = False, daggered: bool = False):
        self.daggered = daggered
        self.U = U
        name_opt = ["initialstate", "simple", "single", "double", "trotter"]
        if name.lower() in name_opt:
            self.name = name.lower()
        else:
            self.name = "simple"
            
        self.qubits = []
        if name == "single" or name == "double":
            for index in U.indices:
                for i in index:
                    self.qubits.append(i)
        elif name == 'trotter':
            self.qubits.extend(U._target)
            self.qubits = list(set(self.qubits))
        else:
            if isinstance(U, tq.QCircuit):
                for gate in self.U.gates:
                    self.qubits.extend(gate.qubits)
                self.qubits = list(set(self.qubits))  # remove duplicates
            elif isinstance(U, tq.circuit._gates_impl.QGateImpl):
                self.qubits.extend(U.qubits)
                self.qubits = list(set(self.qubits))  # remove duplicates
        if not n_qubits_is_double:
            spatial = [q // 2 for q in self.qubits]  # half the number of qubits visualized
            self.qubits = list(set(spatial))
    def __str__(self):
        res = type(self).__name__
        res += ' ' + self.name.title()
        circ = ' with gates: '
        for gate in self.U.gates:
            circ += f'{gate.name }{gate.qubits}'
        res += circ
        return res
    def construct_circuit(self):
        return self.U

    # TODO: this was copied, but can't be right
    def map_variables(self, variables):
        mapped_gate = copy.deepcopy(self)
        for g in mapped_gate.construct_circuit().gates:
            g = g.map_variables(variables)
        return mapped_gate
    
    def used_wires(self) -> List[int]:
        return self.qubits

    def dagger(self) -> "Gate":
        daggered = copy.deepcopy(self)
        daggered.U = daggered.U.dagger()
        daggered.daggered = not daggered.daggered
        return daggered
        

    def _render(self, state: CircuitState, style: CircuitStyle) -> str:
        result = ""
        if self.name != 'trotter':
            for q in self.qubits:
                result += " a{qubit} ".format(qubit=q)
        if self.name == "initialstate":
            result += " G I "
        if self.name == "simple":
            result += " G G "
        if self.name == "single":
            result += " color=blue "
        if self.name == "double":
            result += " color=fai "
        if self.name == "trotter":
            text = tq.circuit.qpic.export_to_qpic(
                circuit=tq.gates.Trotterized(generator=self.U.generator, angle=self.U._parameter))
            while text[0:len("COLOR")] == "COLOR":
                text = text[text.find("\n") + 1:]
            while text[0] == "a":
                text = text[text.find("\n") + 1:]
            result += text

        if self.daggered:
            result += " { $\\dagger$ } "

        return result


from itertools import combinations
from .fermionic_utils import make_excitation_generator_op

def create_fermionic_generators(instructions, angles, form: str = 'fermionic') -> dict:

    #todo check for instruction format: even, repeating indeces

    generators={}
    new_instructions = []

    for exct in instructions:
        aux =[]
        for indicies in exct:
            aux.append(indicies[0])
            aux.append(indicies[1])
        aux = [tuple(aux)]
        new_instructions.append(aux)
    instructions = new_instructions


    for angle_idx, fermionic_circuit in enumerate(instructions):
        # todo checks
        if angles[angle_idx] in generators:
            generators[angles[angle_idx]] = generators[angles[angle_idx]] + make_excitation_generator_op(indices=fermionic_circuit[0], form=form)
        else:
            generators[angles[angle_idx]] = make_excitation_generator_op(indices=fermionic_circuit[0], form=form)

    return generators


def generate_of_binary_dict(n_orb: int, n_e: int) -> dict:

    """
    create a dictionary of all possibilities of a binary string of length n_orb with n_e number of 1s as a key and
    the integer value of the binary string as the value
    Parameters
    ----------
    n_orb: number of orbitals
    n_e: number of electrons

    Returns
    -------
    Dictionary with the key being the binary possibilities and the value being the integer value of the binary string
    """
    result = {}
    for index, positions in enumerate(combinations(range(n_orb), n_e)):
        s = ['0'] * n_orb
        for pos in positions:
            s[pos] = '1'
        binary_str = ''.join(s)

        result[binary_str[::-1]] = index
    return result





import subprocess
import os

def qpic_to_pdf(filename, filepath=None):
    """Calls qpic to generate a pdf from the given apic file."""
    if filepath:
        subprocess.call(["qpic", "{}.qpic".format(filename), "-f", "pdf"], cwd=filepath)
    else:
        subprocess.call(["qpic", "{}.qpic".format(filename), "-f", "pdf"],cwd=os.getcwd())


def qpic_to_png(filename, filepath=None):
    """Calls qpic to generate a png from the given apic file."""

    if filepath:
        subprocess.call(["qpic", "{}.qpic".format(filename), "-f", "png"], cwd=filepath)
    else:
        subprocess.call(["qpic", "{}.qpic".format(filename), "-f", "png"],cwd=os.getcwd())




# Generalized Adaptive Solvers
# as described in Kottmann, Anand, Aspuru-Guzik: https://doi.org/10.1039/D0SC06627C

from tequila import (
    TequilaWarning,
    TequilaException,
)
import numpy
import dataclasses
import warnings
from itertools import combinations
from tequila.apps.adapt.adapt import AdaptParameters
from sunrise.expval import Braket as ExpectationValue
from openfermion.ops.operators.fermion_operator import FermionOperator
from sunrise import (
    grad,
    minimize,
    simulate,
    FCircuit,
)


class AdaptPoolBase:
    """
    Standard class for operator pools in Adapt
    The pool is a list of generators (Excitation Indices [(i,j),(k,l),...])
    """

    generators: list = None

    __n: int = 0  # for iterator, don't touch

    def __init__(self, generators):
        self.generators = generators

    def make_unitary(self, k, label) -> FCircuit:
        return 

    def __iter__(self):
        self.__n = 0
        return self

    def __next__(self):
        if self.__n < len(self.generators):
            result = self.__n
            self.__n += 1
            return result
        else:
            raise StopIteration

    def __str__(self):
        return "{} with {} Generators".format(type(self).__name__, len(self.generators))


class ObjectiveFactoryBase:
    """
    Default class to create the objective in the Adapt solver
    This just creates the single ExpectationValue <H>_{Upre + U + Upost}
    and U will be the circuit that is adaptively constructed
    """

    Upre: FCircuit = FCircuit()
    Upost: FCircuit = FCircuit()
    H: FermionOperator = None

    def __init__(self, H=None, Upre=None, Upost=None,backend=None,molecule=None,*args, **kwargs):
        if molecule is None:
            raise TequilaException("No Molecule was given to Adapt!")

        self.H = H
        if Upre is not None:
            self.Upre = Upre
        else:
            self.Upre = molecule.prepare_reference()
        if Upost is not None:
            self.Upost = Upost
        else:
            self.Upost = FCircuit()
        self.backend=backend
        self.molecule = molecule

    def __call__(self, U, screening=False, *args, **kwargs):
        return ExpectationValue(H=self.H, U=self.Upre + U + self.Upost,backend=self.backend,molecule=self.molecule,*args, **kwargs)

    def grad_objective(self, *args, **kwargs):
        return self(*args, **kwargs)

    def __str__(self):
        return "{}".format(type(self).__name__)


class Adapt:
    operator_pool: AdaptPoolBase = None
    objective_factory = None
    parameters: AdaptParameters = AdaptParameters()

    def make_objective(self, U, variables=None, *args, **kwargs):
        return self.objective_factory(U=U, variables=variables, *args, **{**self.parameters.compile_args, **kwargs})

    def __init__(self, operator_pool, H=None, objective_factory=None,backend=None, *args, **kwargs):
        """
        For the Default Adaptive Solver kwargs can contain Upre and Upost as described in:
        See out online tutorial for more information: https://github.com/tequilahub/tequila-tutorials #TODO: Not exactly tequila since fermionic
        """
        self.operator_pool = operator_pool
        self.backend = backend
        if objective_factory is None:
            self.objective_factory = ObjectiveFactoryBase(H,backend=self.backend, *args, **kwargs)
        else:
            self.objective_factory = objective_factory
        filtered = {k: v for k, v in kwargs.items() if k in self.parameters.__dict__}
        self.parameters = AdaptParameters(*args, **filtered)
        if (
            self.parameters.silent
            and self.parameters.optimizer_args is not None
            and "silent" not in self.parameters.optimizer_args
        ):
            self.parameters.optimizer_args["silent"] = True

    def __call__(self, static_variables=None, mp_pool=None, label=None, variables=None, *args, **kwargs):
        if not self.parameters.silent:
            print("Starting Adaptive Solver")
            print(self)

        # count resources
        screening_cycles = 0
        objective_expval_evaluations = 0
        gradient_expval_evaluations = 0
        histories = []

        if static_variables is None:
            static_variables = {}

        if variables is None:
            variables = {**static_variables}
        else:
            variables = {**variables, **static_variables}

        U = FCircuit()
        if "U" in kwargs:
            U = kwargs["U"]
            kwargs.pop('U')
        elif 'initial_state' in kwargs:
            U.initial_state = kwargs['initial_state']
            kwargs.pop('initial_state')
        elif hasattr(self.operator_pool, "initialize_circuit"):
            U = self.operator_pool.initialize_circuit()
        initial_objective = self.make_objective(U, variables=variables)
        for k in initial_objective.extract_variables():
            if k not in variables:
                warnings.warn(
                    "variable {} of initial objective not given, setting to 0.0 and activate optimization".format(k),
                    TequilaWarning,
                )
                variables[k] = 0.0

        if len(initial_objective.extract_variables()) > 0:
            active_variables = [k for k in variables if k not in static_variables]
            if len(active_variables) > 0:
                if not self.parameters.silent:
                    print("initial optimization")
                margs = {"initial_values": variables}
                margs = {**margs, **self.parameters.compile_args, **self.parameters.optimizer_args}
                result = minimize(objective=initial_objective, variables=active_variables, **margs)

                variables = result.variables
        energy = simulate(initial_objective, variables=variables)
        for iter in range(self.parameters.maxiter):
            current_label = (iter, 0)
            if label is not None:
                current_label = (iter, label)

            gradients = self.screen_gradients(U=U, variables=variables, mp_pool=mp_pool)
            grad_values = numpy.asarray(list(gradients.values()))
            max_grad = max(grad_values)
            grad_norm = numpy.linalg.norm(grad_values)

            if grad_norm < self.parameters.gradient_convergence:
                if not self.parameters.silent:
                    print("pool gradient norm is {:+2.8f}, convergence criterion met".format(grad_norm))
                break
            if numpy.abs(max_grad) < self.parameters.max_gradient_convergence:
                if not self.parameters.silent:
                    print(
                        "max pool gradient is {:+2.8f}, convergence criterion |max(grad)|<{} met".format(
                            max_grad, self.parameters.max_gradient_convergence
                        )
                    )
                break

            batch_size = self.parameters.batch_size

            # detect degeneracies
            degeneracies = [
                k
                for k in range(batch_size, len(grad_values))
                if numpy.isclose(grad_values[batch_size - 1], grad_values[k], rtol=self.parameters.degeneracy_threshold)
            ]

            if len(degeneracies) > 0:
                batch_size += len(degeneracies)
                if not self.parameters.silent:
                    print(
                        "detected degeneracies: increasing batch size temporarily from {} to {}".format(
                            self.parameters.batch_size, batch_size
                        )
                    )

            count = 0

            op_names = []
            for k, v in gradients.items():
                Ux = self.operator_pool.make_unitary(k, label=current_label)
                U += Ux
                op_names.append(Ux.extract_variables())
                count += 1
                if count >= batch_size:
                    break
            variables = {**variables, **{k: 0.0 for k in U.extract_variables() if k not in variables}}
            active_variables = [k for k in variables if k not in static_variables]

            objective = self.make_objective(U, variables=variables)
            margs = {"initial_values": variables}
            margs = {**margs, **self.parameters.compile_args, **self.parameters.optimizer_args}
            result = minimize(objective=objective, variables=active_variables, **margs)
            niter = len(result.history.energies)
            diff = energy - result.energy
            energy = result.energy
            variables = result.variables
            
            if not self.parameters.silent:
                print("-------------------------------------")
                print("Finished iteration {}".format(iter))
                print("added ", op_names)
                print("current energy : {:+2.8f}".format(energy))
                print("difference     : {:+2.8f}".format(diff))
                print("grad_norm      : {:+2.8f}".format(grad_norm))
                print("max_grad       : {:+2.8f}".format(max_grad))
                print("ops in circuis : {}".format(len(U.gates)))
                print("optimizer      : {}".format(self.parameters.optimizer_args["method"]))
                print("opt-iterations : {}".format(niter))

            screening_cycles += 1
            mini_iter = len(result.history.extract_energies())
            gradient_expval = sum([v.count_expectationvalues() for k, v in grad(objective).items()])
            # objective_expval_evaluations += mini_iter * objective.count_expectationvalues() 
            gradient_expval_evaluations += mini_iter * gradient_expval
            histories.append(result.history)

            if self.parameters.energy_convergence is not None and numpy.abs(diff) < self.parameters.energy_convergence:
                if not self.parameters.silent:
                    print("energy difference is {:+2.8f}, convergence criterion met".format(diff))
                break

            if iter == self.parameters.maxiter - 1:
                if not self.parameters.silent:
                    print("reached maximum number of iterations")
                break

        @dataclasses.dataclass
        class AdaptReturn:
            U: FCircuit = None
            objective_factory: ObjectiveFactoryBase = None
            variables: dict = None
            energy: float = None
            histories: list = None
            screening_cycles: int = None
            objective_expval_evaluations: int = None
            gradient_expval_evaluations: int = None

        return AdaptReturn(
            U=U,
            variables=variables,
            objective_factory=self.objective_factory,
            energy=energy,
            histories=histories,
            screening_cycles=screening_cycles,
            objective_expval_evaluations=objective_expval_evaluations,
            gradient_expval_evaluations=gradient_expval_evaluations,
        )

    def screen_gradients(self, U, variables, mp_pool=None):
        args = []
        for k in self.operator_pool:
            arg = {}
            arg["k"] = k
            arg["variables"] = variables
            arg["U"] = U
            args.append(arg)

        if mp_pool is None:
            dEs = [self.do_screening(arg) for arg in args]
        else:
            if not self.parameters.silent:
                print("screen with {} workers".format(mp_pool._processes))
            dEs = mp_pool.map(self.do_screening, args)
        dEs = dict(sorted(dEs, reverse=True, key=lambda x: numpy.fabs(x[1])))
        return dEs

    def do_screening(self, arg):
        Ux = self.operator_pool.make_unitary(k=arg["k"], label="tmp")
        Utmp = arg["U"] + Ux
        variables = {**arg["variables"]}
        objective = self.make_objective(Utmp, screening=True, variables=variables)

        dEs = []
        for k in Ux.extract_variables():
            variables[k] = 0.0
            dEs.append(grad(objective, k))
        gradients = [
            numpy.abs(simulate(objective=dE, variables=variables, **self.parameters.compile_args)) for dE in dEs
        ]
        return arg["k"], sum(gradients)

    def __str__(self):
        result = str(self.parameters)
        result += str("{:30} : {}\n".format("operator pool: ", self.operator_pool))
        result += str("{:30} : {}\n".format("objective factory : ", self.objective_factory))
        return result


class MolecularPool(AdaptPoolBase):
    def __init__(self, molecule, indices: str):
        """

        Parameters
        ----------
        molecule:
            a sunrise fermionic molecule object
        indices
            a list of indices defining UCC operations
            indices refer to spin-orbitals
            e.g. indices = [[(0,2),(1,3)], [(0,2)], [(1,3)]]
            can be a string for predefined pools supported are UpCCD, UpCCSD, UpCCGD, and UpCCGSD
        """
        self.molecule = molecule

        if isinstance(indices, str):
            if "CC" not in indices.upper():
                raise TequilaException(
                    "Pool of type {} not yet supported.\nCreate your own by passing the initialized indices".format(
                        indices
                    )
                )

            generalized = True if "G" in indices.upper() else False
            paired = True if "P" in indices.upper() else False
            singles = True if "S" in indices.upper() else False
            doubles = True if "D" in indices.upper() else False

            indices = []
            if doubles:
                indices += self.make_indices_doubles(generalized=generalized, paired=paired)
            if singles:
                indices += self.make_indices_singles(generalized=generalized)

        indices = [tuple(k) for k in indices]
        super().__init__(generators=indices)

    def make_indices_singles(self, generalized=False):
        indices = []
        for p in range(self.molecule.n_electrons // 2):
            for q in range(self.molecule.n_electrons // 2, self.molecule.n_orbitals):
                indices.append([(2 * p, 2 * q)])
                indices.append([(2 * p + 1, 2 * q + 1)])
        if not generalized:
            return indices

        for p in range(self.molecule.n_orbitals):
            for q in range(p + 1, self.molecule.n_orbitals):
                if [(2 * p, 2 * q)] in indices:
                    continue
                indices.append([(2 * p, 2 * q)])
                indices.append([(2 * p + 1, 2 * q + 1)])
        return self.sort_and_filter_unique_indices(indices)

    def make_indices_doubles(self, generalized=False, paired=True):
        indices = []
        for p in range(self.molecule.n_electrons // 2):
            for q in range(self.molecule.n_electrons // 2, self.molecule.n_orbitals):
                indices.append([(2 * p, 2 * q), (2 * p + 1, 2 * q + 1)])

        if not generalized:
            return indices

        for p in range(self.molecule.n_orbitals):
            for q in range(p + 1, self.molecule.n_orbitals):
                idx = [(2 * p, 2 * q), (2 * p + 1, 2 * q + 1)]
                if idx in indices:
                    continue
                indices.append(idx)

        if not paired:
            indices += self.make_indices_doubles_all(generalized=generalized)

        return self.sort_and_filter_unique_indices(indices)

    def make_indices_doubles_all(self, generalized=False):
        singles = self.make_indices_singles(generalized=generalized)
        unwrapped = [x[0] for x in singles]
        # now make all combinations of singles
        indices = [x for x in combinations(unwrapped, 2)]
        return self.sort_and_filter_unique_indices(indices)

    def sort_and_filter_unique_indices(self, indices):
        # sort as: [[(a,b),(c,d),(e,f)...],...]with a<c, a<b, c<d
        sorted_indices = []
        for idx in indices:
            idx = tuple([tuple(sorted(pair)) for pair in idx])  # sort internal pairs (a<b, c<d, etc)
            # avoid having orbitals show up multiple times in excitatin strings
            idx = tuple(
                [pair for pair in idx if sum([1 for pair2 in idx if pair[0] in pair2 or pair[1] in pair2]) == 1]
            )
            if len(idx) == 0:
                continue
            idx = tuple(list(set(idx)))  # avoid repetitions (like ((0,2),(0,2)))
            idx = tuple(sorted(idx, key=lambda x: x[0]))  # sort pairs by first entry (a<c)
            sorted_indices.append(idx)
        return list(set(sorted_indices))

    def make_unitary(self, k, label):
        return self.molecule.make_excitation_gate(
            indices=self.generators[k], angle=(self.generators[k], label),
        )
    def initialize_circuit(self):
        return self.molecule.prepare_reference()

class PseudoSingletMolecularPool(MolecularPool):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        indices = []
        for idx in self.generators:
            if len(idx) == 1:
                combined = (
                    ((idx[0][0] // 2 * 2, idx[0][1] // 2 * 2)),
                    ((idx[0][0] // 2 * 2 + 1, idx[0][1] // 2 * 2 + 1)),
                )
                if combined not in indices:
                    indices.append(combined)
            else:
                indices.append(tuple([idx]))

        self.generators = list(set(indices))
    def make_unitary(self, k, label):
        U = FCircuit()
        for idx in self.generators[k]:
            combined_variable = self.generators[k][0]
            U += self.molecule.make_excitation_gate(indices=idx, angle=(combined_variable, label))
        return U
    def initialize_circuit(self):
        return self.molecule.prepare_reference()

class ObjectiveFactorySequentialExcitedState(ObjectiveFactoryBase):
    def __init__(self, H, circuits: list, factors: list, *args, **kwargs):
        self.circuits = circuits
        self.factors = factors
        super().__init__(H=H, *args, **kwargs)

    def __call__(self, U, *args, **kwargs):
        circuit = self.Upre + U + self.Upost
        objective = ExpectationValue(H=self.H, U=circuit)
        Qp = FermionOperator(' '.join('{p} {p}^'.format(p=p) for p in U.qubits))
        # get all overlaps  
        for i, Ux in enumerate(self.circuits):
            S2 = ExpectationValue(H=Qp, U=circuit + Ux.dagger())
            objective += numpy.abs(self.factors[i]) * S2
        return objective
    def initialize_circuit(self):
        return self.molecule.prepare_reference()

def run_molecular_adapt(molecule, operator_pool: str = None, Upre=None, Upost=None,backend=None, *args, **kwargs):
    if operator_pool is None:
        operator_pool = "UCCGSD"

    # auto-detect if we have an molecular pool
    # initialized by keyword
    # e.g. U(p)CC(G)(S)(D)
    ucc_signals = ["u", "cc", "s", "d", "g"]
    if hasattr(operator_pool, "lower"):
        if any([s in operator_pool.lower() for s in ucc_signals]):
            operator_pool = MolecularPool(molecule=molecule, indices=operator_pool)

    if Upre is None:
        Upre = molecule.prepare_reference()

    solver = Adapt(operator_pool=operator_pool, Upre=Upre, Upost=Upost,backend=backend, *args, **kwargs)

    result = solver()

    return result


import numbers
from typing import List

from ..core import CircuitState, CircuitStyle
from ..quantum_chemistry import PairCorrelatorGate
from ..core import Polygon
import numbers
from abc import ABC
from math import pi
from typing import List

from ..core import Color
from ..core import CircuitStyle
from ..core import Gate
import tequila as tq
import numpy
import copy

class GenericExcitation(Gate, ABC):
    """
            indices [(i,j),(k,l),...] correspond to the n electron excitation i -> j, k -> l,...
    """

    def __init__(self, indices, angle, control=None, assume_real=True, encoding="JordanWigner",n_qubits_is_double:bool=False, unit_of_pi: bool = False,*args, **kwargs):
        assert isinstance(indices,(list,tuple))
        if isinstance(indices[0],numbers.Number):
            indices = [indices]
        self.indices = indices
        self.idx = []
        for i in self.indices:
            self.idx.extend(i)
        self.angle = tq.assign_variable(angle)
        self.unit_of_pi = unit_of_pi
        self.control = control
        self.assume_real = assume_real
        self.n_qubits_is_double = n_qubits_is_double
        if "molecule" not in kwargs:
            k = max(self.idx)
            x = numpy.zeros(k ** 2).reshape([k, k])
            y = numpy.zeros(k ** 4).reshape([k, k, k, k])
            if encoding is None:
                encoding = "jordanwigner"
            self.encoding = encoding.lower()
            if "molecule_factory" not in kwargs:
                self._dummy = tq.Molecule(geometry="", one_body_integrals=x, two_body_integrals=y,
                                          nuclear_repulsion=0.0,
                                          transformation=self.encoding, *args, **kwargs)
            else:
                mol_fact = kwargs["molecule_factory"]
                kwargs.pop("molecule_factory")
                self._dummy = mol_fact(geometry="", one_body_integrals=x, two_body_integrals=y, nuclear_repulsion=0.0,
                                       transformation=self.encoding, *args, **kwargs)
        else:
            self._dummy = kwargs["molecule"]
            self.encoding = self._dummy.transformation
            kwargs.pop("molecule")

    def construct_circuit(self, *args, **kwargs):
        return self._dummy.make_excitation_gate(angle=self.angle, control=self.control,indices=self.indices,assume_real=self.assume_real)

    def map_variables(self, variables):
        mapped_gate = copy.deepcopy(self)
        mapped_gate.angle = mapped_gate.angle.map_variables(variables)
        return mapped_gate

    def used_wires(self) -> List[int]:

        used = list(range(min(self.idx) // (1+(not self.n_qubits_is_double)), max(self.idx) // (1+(not self.n_qubits_is_double)) + 1))
        if self.control is not None:
            used.append(self.control // (1+(not self.n_qubits_is_double)))

        return used

    def dagger(self) -> "Gate":
        raise NotImplemented

    def render_angle(self, style: CircuitStyle, gcol: Color, tcol: Color) -> [Color, Color]:
        if not isinstance(self.angle, numbers.Number) and style.parametrized_marking is not None:
            tcol = style.parametrized_marking.text_color
            gcol = style.parametrized_marking.gate_color

        if isinstance(self.angle, numbers.Number) and style.color_range is not None:
            if self.unit_of_pi:
                angle = int(((self.angle % 2) / 2) * 100)
            else:
                angle = int(abs((self.angle / (2 * pi)) * 100)) % 100
            gcol = style.color_range.interpolate(angle)

        return gcol, tcol

    def _render(self, state, style):
        spin = {0: Polygon(-3), 1: Polygon(3)}
        # styling
        tcol = style.primary.text_color
        gcol = style.primary.gate_color

        # angle
        gcol, tcol = self.render_angle(style, gcol, tcol)
        if 'generic' in style.personalized:
            gcol = style.personalized['generic'].gate_color
            tcol = style.personalized['generic'].text_color
        param = self.angle
        if isinstance(param, numbers.Number):
            param = "{:1.2f}".format(param)

        # rendering
        result = ""
        for exct in self.indices:
            shape1 = spin[exct[0] % 2]
            shape2 = spin[exct[1] % 2]
            result += " a{qubit} P:fill={gcol}:shape={shape} \\textcolor{tcol}{{{op}}} ".format(qubit=exct[0] // (1+(not self.n_qubits_is_double)),shape=shape1,gcol=gcol,tcol="{" + tcol.name + "}",op="")
            result += " a{qubit} P:fill={gcol}:shape={shape} \\textcolor{tcol}{{{op}}} ".format(qubit=exct[1] // (1+(not self.n_qubits_is_double)),shape=shape2,gcol=gcol,tcol="{" + tcol.name + "}",op="")

        return result


import typing
import numbers
from tequila.objective.objective import FixedVariable,Variable
from .circuit import FCircuit
from .fgateimpl import *
from .givens_rotations import get_givens_circuit
from numpy import ndarray

def FermionicExcitation(indices:typing.Union[list,tuple], variables:typing.Union[typing.Hashable, numbers.Real, Variable, FixedVariable]=None, reordered:bool=False)->FCircuit:
    return FCircuit.wrap_gate(FermionicExcitationImpl(indices,variables,reordered))

def UR(i:int,j:int, variables:typing.Union[typing.Hashable, numbers.Real, Variable, FixedVariable]=None)->FCircuit:
    return FCircuit.wrap_gate(URImpl(i,j,variables))

def UC(i:int,j:int, variables:typing.Union[typing.Hashable, numbers.Real, Variable, FixedVariable]=None)->FCircuit:
    return FCircuit.wrap_gate(UCImpl(i,j,variables))

def UX(indices:typing.Union[list,tuple], variables:typing.Union[typing.Hashable, numbers.Real, Variable, FixedVariable]=None, reordered:bool=False)->FCircuit:
    return FCircuit.wrap_gate(FermionicExcitationImpl(indices,variables,reordered))

def Phase(tarjet:int,variables:typing.Union[typing.Hashable, numbers.Real, Variable, FixedVariable]=None, reordered:bool=False)->FCircuit:
    return FCircuit.wrap_gate(PhaseImpl(tarjet,variables,reordered))

def Givens(unitary:ndarray, tol:float=1e-12, ordering:typing.Union[list,tuple,str]='OPTIMIZED_ORDERING')->FCircuit:
    return get_givens_circuit(unitary,tol,ordering)


import abc

class Shape(abc.ABC):
    """Abstract base class for all qpic shapes."""
    @abc.abstractmethod
    def to_qpic(self) -> str:
        """Returns the qpic string identifying this shape."""
        pass

    def __format__(self, format_spec) -> str:
        return self.to_qpic()


class NoBoundary(Shape):
    """This draws a gate without a boundary."""
    def to_qpic(self):
        return "0"

class Control(Shape):
    """This draws the gate as a control. It can optionally be drawn as a negated control."""

    def __init__(self, negated: bool = False):
        self.negated = negated

    def to_qpic(self):
        if self.negated:
            return "-1"
        else:
            return "1"

class Circle(Shape):
    """This draws a circle."""
    def to_qpic(self):
        return "2"

class Polygon(Shape):
    """
    This draws any polygon with n >= 3 vertices.
    If bottom_vertex is true, odd numbered polygons will draw their tip at the bottom instead of at the top. (Basically rotating by 180°)
    """
    def __init__(self, vertices: int, bottom_vertex: bool = False):
        if vertices < 0:
            vertices = vertices * -1
            bottom_vertex = not bottom_vertex

        assert vertices >= 3
        self.vertices = vertices
        self.bottom_vertex = bottom_vertex

    def to_qpic(self):
        if self.bottom_vertex:
            return str(-self.vertices)
        else:
            return str(self.vertices)

class Triangle(Shape):
    """This draws a triangle pointing to the right, or left when backward is True."""
    def __init__(self, backward: bool = False):
        self.backward = backward

    def to_qpic(self):
        if self.backward:
            return "<"
        else:
            return ">"


from optparse import Option
from typing import List, Union, Optional,Dict

from .color import ColorRange, Colors, Color


class CircuitStyle:
    """
    This class defines the style that should be used when rendering the circuit.
    Parameters:
        group_together -- If true, this touches all wires after each gate, preventing gates from being rearranged.
        parametrized_marking -- If set the given Marking is used to color parametrized gates.
        color_range -- If set the given ColorRange is used to color parametrized gates according to their angle.

        primary, secondary -- Basic colors to be used by the gates, each contains a text and gate color.
    """

    def __init__(self,
                 group_together: bool = False,
                 parametrized_marking: Optional[Colors] = Colors(Color("black"), Color("unia")),
                 color_range: Optional[ColorRange] = ColorRange(Color("blue"), Color("red")),
                 primary: Colors = Colors(Color("white"), Color("tq")),
                 secondary: Colors = Colors(Color("white"), Color("fai")),
                 personalized:Dict['str',Colors]={}):
        self.group_together = group_together
        self.parametrized_marking = parametrized_marking
        self.color_range = color_range
        self.primary = primary
        self.secondary = secondary
        self.personalized = personalized


class CircuitState:
    """
    This class contains the current state of the circuit.
    This is mostly needed for swap networks, to store the order of the wires.
    It also stores the number of orbitals.
    """

    def __init__(self, n_orbitals: int):
        self.n_orbitals = n_orbitals
        self.wires = list(range(n_orbitals))

    @staticmethod
    def names(wires: Union[List[int], int]) -> str:
        """Generates the name of a single or multiple wires by their indices."""
        if wires is int:
            wires = [wires]

        return " ".join(f"a{i}" for i in wires)

    def all_names(self) -> str:
        """Generates names for all wires known to this state"""
        return self.names(self.wires)



from sunrise import molecules
from sunrise.molecules import Molecule
from sunrise.plot_MO import plot_MO
from sunrise.graphical.qpic_visualization import qpic_to_pdf,qpic_to_png
from sunrise import measurement
from sunrise.fermionic_operations.orb_rotation_qubit import OrbitalRotation
from sunrise.miscellaneous.giuseppe import giuseppe
from sunrise.miscellaneous.bar import giuseppe_bar
from sunrise.expval import Braket,show_available_modules,show_supported_modules
from sunrise.fermionic_operations import gates
from sunrise.fermionic_operations.circuit import FCircuit
from sunrise.expval.pyscf_molecule import *
from sunrise.expval.minimize import grad,minimize,simulate
from sunrise.expval.optimize import optimize_orbitals
from sunrise import graphical
from sunrise.expval import Braket as Expval


import numpy as np
import tequila as tq
import pytest
import sunrise as sn
from sunrise.expval import INSTALLED_FERMIONIC_BACKENDS,Braket
from numpy import isclose
import random
from datetime import datetime



HAS_TCC = "tcc" in INSTALLED_FERMIONIC_BACKENDS
HAS_FQE = "fqe" in INSTALLED_FERMIONIC_BACKENDS

@pytest.mark.parametrize("geom",["H 0.0 0.0 0.0\nH 0.0 0.0 1.6\nH 0.0 0.0 3.2\nH 0.0 0.0 4.8","H 0. 0. 0.\n Be 0. 0. 1.6\n H 0. 0. 3.2"])
@pytest.mark.parametrize('backend',INSTALLED_FERMIONIC_BACKENDS)
def test_spa(geom,backend):
    mol = tq.Molecule(geometry=geom,basis_set='sto-3g',transformation='reordered-jordan-wigner').use_native_orbitals()
    edges = sn.Molecule(geometry=geom,basis_set='sto-3g',nature='hybrid').get_spa_edges()
    U = mol.make_ansatz("SPA",edges=edges,optimize=False)
    circuit = sn.FCircuit.from_edges(edges=edges,n_orb=mol.n_orbitals)
    expval = tq.ExpectationValue(H=mol.make_hamiltonian(),U=U)
    sunval = Braket(molecule=mol,ket=circuit,backend=backend)
    tqE = tq.minimize(expval,silent=True)
    e = sn.minimize(sunval, silent=True)
    sunE = e.energy
    tqwfn = tq.simulate(U,tqE.angles)
    sunwfn = sn.simulate(U,e.variables)
    assert isclose(tqE.energy,sunE)
    assert isclose(abs(tqwfn.inner(sunwfn)),1,1.e-3)


@pytest.mark.parametrize("geom",["H 0.0 0.0 0.0\nH 0.0 0.0 1.6\nH 0.0 0.0 3.2\nH 0.0 0.0 4.8","H 0. 0. 0.\n Be 0. 0. 1.6\n H 0. 0. 3.2"])
@pytest.mark.parametrize('backend',INSTALLED_FERMIONIC_BACKENDS)
def test_upccsd(geom,backend):
    mol = tq.Molecule(geometry=geom,basis_set='sto-3g',transformation='reordered-jordan-wigner')
    U = mol.make_ansatz("UpCCSD")
    fmol = sn.Molecule(geometry=geom,basis_set='sto-3g',nature='fermionic')
    circuit = fmol.make_ansatz("UpCCSD")
    expval = tq.ExpectationValue(H=mol.make_hamiltonian(),U=U)
    sunval = Braket(molecule=mol,circuit=circuit,backend=backend)
    tqE = tq.minimize(expval,silent=True)
    snE = sn.minimize(sunval, silent=True)
    assert isclose(tqE.energy,snE.energy)

@pytest.mark.parametrize('backend',INSTALLED_FERMIONIC_BACKENDS)
def test_transition(backend):
    geom = 'H 0. 0. 0. \n H 0. 0. 1. \n H 0. 0. 2. \n H 0. 0. 3.'
    mol = tq.Molecule(geometry=geom,basis_set='sto-3g',transformation='reordered-jordan-wigner').use_native_orbitals()
    H = mol.make_hamiltonian()
    U1 = mol.make_ansatz("SPA",edges=[(0,1),(2,3)])
    U2 = mol.make_ansatz("SPA",edges=[(0,2),(1,3)])
    res1 = tq.minimize(tq.ExpectationValue(H=H,U=U1),silent=True)
    res2 = tq.minimize(tq.ExpectationValue(H=H,U=U2),silent=True)
    bra = sn.FCircuit.from_edges([(0,1),(2,3)],n_orb=mol.n_orbitals)
    ket = sn.FCircuit.from_edges([(0,2),(1,3)],n_orb=mol.n_orbitals)
    rov,iov = tq.BraKet(bra=U1,ket=U2,H=H)
    res1.angles.update(res2.angles)
    tq_ov = tq.simulate(rov,variables=res1.angles) + tq.simulate(iov,variables=res1.angles)
    sn_ov = sn.simulate(Braket(molecule=mol,bra=bra,ket=ket,backend=backend),variables=res1.angles)
    assert isclose(tq_ov,sn_ov,atol=1.e-3)

@pytest.mark.parametrize("geom",["H 0.0 0.0 0.0\nH 0.0 0.0 1.6\nH 0.0 0.0 3.2\nH 0.0 0.0 4.8","H 0. 0. 0.\n Be 0. 0. 1.6\n H 0. 0. 3.2"])
@pytest.mark.parametrize('backend',INSTALLED_FERMIONIC_BACKENDS)
def test_maped_variables(geom,backend):
    random.seed(datetime.now().timestamp())
    mol = tq.Molecule(geometry=geom,basis_set='sto-3g',transformation='reordered-jordan-wigner').use_native_orbitals()
    edges = sn.Molecule(geometry=geom,basis_set='sto-3g',nature='hybrid').get_spa_edges()
    U = mol.make_ansatz("SPA",edges=edges,optimize=backend!='tequila')
    mapa = {d:random.random()*np.pi for d in U.extract_variables()}
    U = U.map_variables(mapa)
    circuit = sn.FCircuit.from_edges(edges=edges,n_orb=mol.n_orbitals)
    circuit = circuit.map_variables(mapa)
    expval = tq.ExpectationValue(H=mol.make_hamiltonian(),U=U)
    sunval = Braket(molecule=mol,ket=circuit,backend=backend)
    tqE = tq.simulate(expval,{})
    sunE = sn.simulate(sunval,{})
    assert isclose(tqE,sunE)


@pytest.mark.parametrize("geom",["H 0.0 0.0 0.0\nH 0.0 0.0 1.6\nH 0.0 0.0 3.2\nH 0.0 0.0 4.8","H 0. 0. 0.\n Be 0. 0. 1.6\n H 0. 0. 3.2"])
@pytest.mark.parametrize('backend',INSTALLED_FERMIONIC_BACKENDS)
def test_optimize_orbitals(geom,backend):
    if backend == "tequila":
        pytest.skip("Tequila backend requires a Qubit-based molecule")
    snmol = sn.Molecule(geometry=geom,basis_set='sto-3g',nature='f').use_native_orbitals()
    edges = snmol.get_spa_edges()
    initial_guess = snmol.get_spa_guess().T
    tqmol = tq.Molecule(geometry=geom,basis_set='sto-3g',transformation='reordered-jordan-wigner').use_native_orbitals()
    snU = snmol.make_ansatz('SPA',edges=edges)
    tqU = tqmol.make_ansatz('HCB-SPA',edges=edges)
    snopt = sn.optimize_orbitals(molecule=snmol,circuit=snU,backend=backend,silent=True,initial_guess=initial_guess)
    tqopt = tq.chemistry.optimize_orbitals(molecule=tqmol,circuit=tqU,use_hcb=True,silent=True,initial_guess=initial_guess)
    assert isclose(snopt.energy,tqopt.energy)

#TODO: recursion limit problem on tequila, return when fixed
# @pytest.mark.parametrize("geom", ["H 0.0 0.0 0.0\nH 0.0 0.0 1.6\nH 0.0 0.0 3.2\nH 0.0 0.0 4.8"])
# @pytest.mark.parametrize('backend',INSTALLED_FERMIONIC_BACKENDS)
# def test_overlap_minimzation(geom,backend):
#     if backend == "tequila":
#         pytest.skip("Skipping tequila")
#     mol = tq.Molecule(geometry=geom, basis_set='sto-3g',transformation='reordered-jordan-wigner').use_native_orbitals()
#     H = mol.make_hamiltonian()
#     U1 = mol.make_ansatz("SPA", edges=[(0, 1), (2, 3)])
#     U2 = mol.make_ansatz("SPA", edges=[(0, 2), (1, 3)])
#     bra = sn.FCircuit.from_edges([(0, 1), (2, 3)], n_orb=mol.n_orbitals)
#     ket = sn.FCircuit.from_edges([(0, 2), (1, 3)], n_orb=mol.n_orbitals)

#     tqS = tq.minimize(tq.BraKet(bra=U1, ket=U2, H=H)[0], silent=True)
#     snS = sn.minimize(sn.Braket(molecule=mol, ket=ket, bra=bra,backend=backend),silent=True)
#     assert isclose(tqS.energy, snS.energy, atol=1.e-3)

@pytest.mark.parametrize("geom",["H 0.0 0.0 0.0\nH 0.0 0.0 1.6\nH 0.0 0.0 3.2\nH 0.0 0.0 4.8","H 0. 0. 0.\n Be 0. 0. 1.6\n H 0. 0. 3.2"])
@pytest.mark.parametrize('backend',INSTALLED_FERMIONIC_BACKENDS)
def test_gradient(geom,backend):
    tqmol = tq.Molecule(geometry=geom,basis_set='sto-3g',transformation='reordered-jordan-wigner',units='a').use_native_orbitals()
    snmol = sn.Molecule(geometry=geom,basis_set='sto-3g',nature='f').use_native_orbitals()
    random.seed(datetime.now().timestamp())
    tqU = tqmol.make_ansatz('UpCCSD',hcb_optimization=backend=='fqe')
    snU = snmol.make_ansatz('UpCCSD')
    tqval = tq.ExpectationValue(H=tqmol.make_hamiltonian(),U=tqU)
    snval = sn.Braket(molecule=snmol,ket=snU,backend=backend)
    n = random.sample(range(0, len(tqval.extract_variables())), 5)
    variables = [tqval.extract_variables()[i] for i in n]
    values = {d:random.random()*np.pi for d in tqval.extract_variables()}
    tqg = [tq.simulate(tq.grad(tqval,v),variables=values) for v in variables]
    sng = [sn.simulate(sn.grad(snval,v),variables=values) for v in variables]
    assert np.allclose(tqg,sng)


import tequila as tq
from tequila import QCircuit,QTensor
from numpy import zeros,eye,allclose,ndarray,array
from typing import Union,List
from tequila.circuit._gates_impl import ParametrizedGateImpl, QGateImpl


def OR(*args, **kwargs):
    gate = OrbitalRotation(*args, **kwargs)
    return tq.QCircuit.wrap_gate(gate)

# gate impl
class OrbitalRotation(QGateImpl):

    def __init__(self,orbitals:List[int]=None,matrix:Union[QTensor,ndarray,list]=None,molecule=None):
        if isinstance(matrix,list):
            matrix = array(list)
        if isinstance(matrix,ndarray) and not isinstance(matrix,QTensor):
            assert allclose(eye(len(matrix)), matrix.dot(matrix.T.conj())) #normalized
            matrix = QTensor(shape=matrix.shape,objective_list=matrix.reshape(matrix.shape[0]*matrix.shape[1]))
        self.orbital = orbitals
        self.coeff = matrix
        if molecule is None:
            geom = ""
            for k in range(2 * (len(self.coeff) // 2 + 1)):
                geom += f"h 0.0 0.0 {1.5 * k}\n"
            dummy = tq.Molecule(geometry=geom, basis_set='sto-3g')
            self.molecule = dummy
        else: self.molecule = molecule
        assert len(self.orbital) == len(self.coeff)
        assert self.coeff.ndim == 2
        assert self.coeff.shape[0] == self.coeff.shape[1] #is square
        taget = []
        for i in self.orbital:
            taget.append(2*i)
            taget.append(2*i+1)
        super().__init__(name="OrbRot", target=taget, control=None)
        # assert allclose(eye(len(self.coeff)), self.coeff.dot(self.coeff.T.conj())) #not normalization can be enforced, let to the user

    def is_parametrized(self) -> bool:
        return True
    def extract_variables(self):
        return self.coeff.extract_variables()

    def __add__(self, other):
        if isinstance(other,QCircuit):
            return self.compile() + other
        else:
            ndx = list(set(self.orbital + other.orbital))
            pos_idx = [ndx.index(i) for i in self.orbital]
            pos_jdx = [ndx.index(i) for i in other.orbital]
            new_a = self.pad_eye(self.coeff, size=len(ndx))
            new_b = self.pad_eye(other.coeff, size=len(ndx))
            ra = self._rot_mat(len(ndx), pos_idx)
            rb = self._rot_mat(len(ndx), pos_jdx)
            ap = ra.T.dot(new_a.dot(ra))
            bp = rb.T.dot(new_b.dot(rb))
            return OrbitalRotation(matrix=bp.dot(ap), orbitals=ndx)
    def __iadd__(self, other):
        ndx = list(set(self.orbital + other.orbital))
        pos_idx = [ndx.index(i) for i in self.orbital]
        pos_jdx = [ndx.index(i) for i in other.orbital]
        new_a = self.pad_eye(self.coeff, size=len(ndx))
        new_b = self.pad_eye(other.coeff, size=len(ndx))
        ra = self._rot_mat(len(ndx), pos_idx)
        rb = self._rot_mat(len(ndx), pos_jdx)
        ap = ra.T.dot(new_a.dot(ra))
        bp = rb.T.dot(new_b.dot(rb))
        self.coeff = bp.dot(ap)
        self.orbital = ndx
        return self
    def __radd__(self, other):
        return self.__add__(other)

    def compile(self, **kwargs)->QCircuit:
        if "tol" in kwargs:
            tol = kwargs['tol']
            kwargs.pop('tol')
        else:
            tol = 1e-12
        if 'ordering' in kwargs:
            ordering = kwargs['ordering']
            kwargs.pop('ordering')
        else:
            ordering = 'Optimized'
        U = self.molecule.get_givens_circuit(unitary=self.coeff,tol=tol,ordering=ordering)
        d = {2*i:2*idx for i,idx in enumerate(self.orbital)}
        d.update({2*i+1:2*idx+1 for i,idx in enumerate(self.orbital)})
        U = U.map_qubits(qubit_map=d)
        return U
    # def __eq__(self, other):
    #     return self.orbital==other.orbital and allclose(self.coeff,other.coeff)
    def _rot_mat(self,n: int, pos_dx):
        m = QTensor(shape=(n,n),objective_list=zeros(shape=(n,n)).reshape(n*n))
        for i in range(len(pos_dx)):
            m[i, pos_dx[i]] = 1
        rest = [i for i in [*range(n)] if i not in pos_dx]
        for i in range(len(rest)):
            m[len(pos_dx) + i, rest[i]] = 1
        return m
    def __str__(self):
        result = 'Orbital Rotation \n'
        result += f'acting on MO: {self.orbital}\n'
        result += f'with matrix: \n {self.coeff}'
        return result
    def __repr__(self):
        return self.__str__()
    def pad_eye(self,matrix:QTensor,size)->QTensor:
        '''
        Increases the square Qtensor size to (size,size) setting new diag entries to one.
        Useful for Qtensor  tensor product.
        '''
        new = QTensor(shape=(size,size),objective_list=eye(size).reshape(size**2))
        for i in range(len(matrix)):
            for j in range(i,len(matrix)):
                new[i,j] = matrix[i,j]
                new[j,i] = matrix[j,i]
        return new